# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import bpy
import re
import sys
import addon_utils
from datetime import datetime
import bpy.utils.previews
import inspect
from bpy.app.handlers import persistent
import requests
#Utils Comment
exceptional_names={'gui':'blender-osm','ape':'blender-osm','mmd_tools_local':'cats-blender-plugin-development'}
item_panels=['VIEW3D_PT_context_properties',]
view_panels=['VIEW3D_PT_view3d_properties','VIEW3D_PT_view3d_lock','VIEW3D_PT_view3d_cursor','VIEW3D_PT_grease_pencil','VIEW3D_PT_collections','NODE_PT_backdrop','NODE_PT_annotation','IMAGE_PT_view_display','IMAGE_PT_uv_cursor','IMAGE_PT_annotation']
tool_panels=['VIEW3D_PT_active_tool','VIEW3D_PT_tools_object_options','VIEW3D_PT_tools_object_options_transform','WORKSPACE_PT_main','WORKSPACE_PT_custom_props','WORKSPACE_PT_addons','NODE_PT_active_tool','IMAGE_PT_active_tool']
options_panels={'NODE_WORLD_PT_viewport_display','NODE_PT_quality','NODE_MATERIAL_PT_viewport','NODE_EEVEE_MATERIAL_PT_settings','NODE_DATA_PT_light','NODE_DATA_PT_EEVEE_light','NODE_CYCLES_WORLD_PT_settings_volume','NODE_CYCLES_WORLD_PT_settings_surface','NODE_CYCLES_WORLD_PT_settings','NODE_CYCLES_WORLD_PT_ray_visibility','NODE_CYCLES_MATERIAL_PT_settings_volume','NODE_CYCLES_MATERIAL_PT_settings_surface','NODE_CYCLES_MATERIAL_PT_settings','NODE_CYCLES_LIGHT_PT_light','NODE_CYCLES_LIGHT_PT_beam_shape'}
node_panels={'TI_OT_Transfer_Panel','BAN_PT_Bake_A_Node','NODE_PT_texture_mapping','NODE_PT_simulation_zone_items','NODE_PT_active_node_properties','NODE_PT_active_node_generic','NODE_PT_active_node_color','NODE_PT_PanelLinkedEdit'}
addons_with_multiple_tabs={'home_builder':'Home Builder','shino':'Shino','cats-blender':'CATS'}
LOGGING=True
def log(*args):
    if LOGGING:
        print(*args)
def preferences():
    return bpy.context.preferences.addons[__package__].preferences
def get_custom_module_name(bl_type,name):
    if preferences().filter_internal_tabs:
        if bl_type.__name__ in view_panels:
            name='View'
        if bl_type.__name__ in item_panels:
            name='Item'
        if bl_type.__name__ in tool_panels:
            name='Tool'
        if bl_type.__name__ in options_panels:
            name='Options'
        if bl_type.__name__ in node_panels:
            name='Node'
    return name
def split_keep_substring(string, separator=',', substrings=('DAZ (.duf, .dsf) importer',)):
    holder=[]
    for sub in substrings:
        if sub in string:
            string=string.replace(sub,'')
            holder.append(sub)
    result=string.split(separator)
    result=[a for a in result if a.strip()]
    return result+holder
regex=re.compile("^\s*bl_category\s*=\s*.+")
def change_category(index,remove=False):
    #directory=os.path.dirname(os.path.dirname(__file__))
    directories=[]
    addons=[]
    a=preferences().addon_info_for_renaming[index]
    
    #print(addons)
    addons=[a.name,]
    new_category=a.tab_name
    #print(a.name,a.tab_name)
    for mod in addon_utils.modules():
        if mod.__name__ in addons:
            version=bpy.app.version
            addon_dir_path=os.path.join(os.path.dirname(bpy.app.binary_path),f"{version[0]}.{version[1]}","scripts","addons")
            directories.append(os.path.dirname(mod.__file__) if os.path.dirname(mod.__file__)!=os.path.dirname(os.path.dirname(__file__)) and os.path.dirname(mod.__file__)!=addon_dir_path else mod.__file__)
            
        else:
            pass
    if not os.path.isdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels")):
        os.mkdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels"))
    for directory in directories:
        if os.path.isfile(directory):
            if not remove and preferences().auto_backup_addons:
                shutil.copy(directory, os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels",os.path.splitext(os.path.basename(directory))[0]+str(datetime.now().strftime(r' %d-%m-%Y %H %M'))+os.path.splitext(os.path.basename(directory))[1]))
                print("Creating Backup....\n", os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels",os.path.splitext(os.path.basename(directory))[0]+str(datetime.now().strftime(r' %d-%m-%Y %H %M'))+os.path.splitext(os.path.basename(directory))[1]))
            f=directory
            #print(f)
            try:
                data=[]
                with open(f,mode='r') as file:
                    replaced=False
                    while True:
                        line=file.readline()
                        if not line:
                            break
                        if remove:
                            if not "#--changed-by-CleanPanels--" in line:
                                replaced=True
                                data.append(line.replace("#--category-editied-by-CleanPanels--",""))
                            
                        else:
                            
                            match=regex.search(line)
                            if  ("bl_category" in line and new_category in line):
                                data.append(line)
                            else:
                                if "#--changed-by-CleanPanels--" not in line:
                                    
                                    if match:
                                        replaced=True
                                        data.append(line.replace("bl_category","#--category-editied-by-CleanPanels--bl_category"))
                                    else:
                                        data.append(line)
                            if match and not  ("bl_category" in line and new_category in line):
                                print("Replacing :",line)
                                replaced=True
                                data.append(line[:line.index("bl_category")]+f"bl_category='{new_category}'#--changed-by-CleanPanels--\n")
                if data:
                    if replaced:
                        with open(f,mode='w') as file:
                                file.writelines(data)
            except Exception as e:
                log("Error in file",f,'\n',e)
        else:
            #print(directory,os.path.dirname(os.path.dirname(__file__)))
            if directory!=os.path.dirname(os.path.dirname(__file__)):
                
                if not remove and preferences().auto_backup_addons:
                    dest_path=os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels",os.path.basename(directory)+str(datetime.now().strftime(r' %d-%m-%Y %H %M')))
                    if not os.path.isdir(dest_path):
                        shutil.copytree(directory, dest_path)
                        print("Creating Backup....\n", dest_path)
                for f in get_all_python_files(directory):
                    #print(f)
                    try:
                        data=[]
                        with open(f,mode='r') as file:
                            replaced=False
                            while True:
                                line=file.readline()
                                if not line:
                                    break
                                if remove:
                                    if not "#--changed-by-CleanPanels--" in line:
                                        replaced=True
                                        data.append(line.replace("#--category-editied-by-CleanPanels--",""))
                                    
                                else:
                                    match=regex.search(line)
                                    if  ("bl_category" in line and new_category in line):
                                        data.append(line)
                                    else:
                                        if "#--changed-by-CleanPanels--" not in line:
                                            
                                            if match:
                                                replaced=True
                                                data.append(line.replace("bl_category","#--category-editied-by-CleanPanels--bl_category"))
                                            else:
                                                data.append(line)
                                    if match and not  ("bl_category" in line and new_category in line):
                                        print("Replacing :",line)
                                        replaced=True
                                        data.append(line[:line.index("bl_category")]+f"bl_category='{new_category}'#--changed-by-CleanPanels--\n")
                        if data:
                            if replaced:
                                with open(f,mode='w') as file:
                                        file.writelines(data)
                    except Exception as e:
                        log("Error in file",f,'\n',e)
def get_installed_addon_names(self, context):
    addons=[]
    for a in bpy.context.preferences.addons.keys():

        try:
            if a not in split_keep_substring(preferences().addons_to_exclude)+addons_to_exclude and a not in split_keep_substring(preferences().workspace_categories[self.index].panels):
                mod = sys.modules[a]
                addons.append(mod.bl_info.get('name', "Unknown"))
        except:
            pass
    addons=sorted(addons,key=str.casefold)
    return addons
def get_all_addon_names(self, context):
    addons=[]
    for a in bpy.context.preferences.addons.keys():

        try:
            if a not in split_keep_substring(preferences().addons_to_exclude)+addons_to_exclude:
                mod = sys.modules[a]
                addons.append(mod.bl_info.get('name', "Unknown"))
        except:
            pass
    addons=sorted(addons,key=str.casefold)
    return addons
def get_all_addons(self, context):
    addons=[]
    for a in bpy.context.preferences.addons.keys():

        try:
            mod = sys.modules[a]
            a=mod.bl_info.get('name', "Unknown")
            if a not in addons_to_exclude+(split_keep_substring(preferences().addons_to_exclude) if preferences().use_enum_search_for_popups else []):
                
                addons.append((a,a))
                # mod = sys.modules[a]
                # if mod not in preferences().addons_to_exclude)+addons_to_exclude:
                #     addons.append((mod.bl_info.get('name', "Unknown"),mod.bl_info.get('description', "")))
        except:
            pass
    internal_tabs=[]
    if preferences().filter_internal_tabs:
        internal_tabs=[("Tool","Tool"),("View","View"),("Node","Node"),("Options","Options")]
    addons=sorted(addons,key=lambda x:x[0].lower())
    return [(a,a+("(Tab)" if a in internal_tabs else ""),a) for a,b in addons]+[("Unfiltered","Unfiltered","Unfiltered"),]
def get_installed_addon_names(self, context):
    addons=[]
    for a in bpy.context.preferences.addons.keys():

        try:
            if a not in split_keep_substring(preferences().addons_to_exclude)+addons_to_exclude and a not in split_keep_substring(preferences().workspace_categories[self.index].panels):
                mod = sys.modules[a]
                addons.append(mod.bl_info.get('name', "Unknown"))
        except:
            pass
    addons=sorted(addons,key=str.casefold)
    return addons
def get_installed_addons(self, context):
    addons=[]
    others=[]
    for i,a in enumerate(preferences().workspace_categories):
        if i!=self.index:
            others.extend(split_keep_substring(preferences().workspace_categories[i].panels))
    others.extend(split_keep_substring(preferences().addons_to_exclude))
    for a in bpy.context.preferences.addons.keys():

        try:
            if a not in ['View','Tool','Node','Options']:
                mod = sys.modules[a]
                a=mod.bl_info.get('name', "Unknown")
            if a not in split_keep_substring(preferences().addons_to_exclude)+addons_to_exclude and a not in (split_keep_substring(preferences().workspace_categories[self.index].panels) if preferences().use_enum_search_for_popups else []):
                addons.append(a)
                #addons.append(a)
        except:
            pass
    # if preferences().filter_internal_tabs:
    #     addons.append("Tool")
    #     addons.append("View")
    #     addons.append("Node")
    #     addons.append("Options")
    internal_tabs=[]
    if preferences().filter_internal_tabs:
        internal_tabs=["Tool","View","Node","Options"]
    addons=sorted(addons,key=str.casefold)
    return [(a,a+("(Tab)" if a in internal_tabs else ""),a,'CHECKMARK' if a in others else 'NONE',i) for i,a in enumerate(addons+internal_tabs)]+[("All",'All','All'),("Unfiltered","Unfiltered","Unfiltered")]
def get_module_name_from_addon_name(name):
    if name in ['Tool','View','Node','Options']:
         return name
    for a in bpy.context.preferences.addons.keys():
        try:
            if name==a:
                return a
            mod = sys.modules[a]
            if mod.bl_info.get('name', "Unknown")==name:
                return a
        except:
            pass
    return "--Unknown--"
def get_all_panel_categories(self, context):
    cat=set()
    base_type = bpy.types.Panel
    for typename in dir(bpy.types):
        
        try:
            bl_type = getattr(bpy.types, typename,None)
            if issubclass(bl_type, base_type):
                if getattr(bl_type,'backup_space',"None")=='VIEW_3D' or getattr(bl_type,'bl_space_type',"None")=='VIEW_3D':
                    if getattr(bl_type,'backup_category',None):
                        cat.add(getattr(bl_type,'backup_category',None))
                    if getattr(bl_type,'bl_category',None):
                        cat.add(getattr(bl_type,'bl_category',None))
                    #cat.add(getattr(bl_type,'backup_category',None) if getattr(bl_type,'backup_category',None) else getattr(bl_type,'bl_category',"None"))
        except:
            pass
    cat=[a for a in cat if a!=preferences().holder_tab_name and a!='None']
    cat=sorted(cat)
    return [(a,a,a) for a in cat]
def get_panel_categories(self, context):
    cat=set()
    base_type = bpy.types.Panel
    others=[]
    if self.category =='FP':
        if preferences().filter_internal_tabs:
                cat.add("Tool")
                cat.add("View")
        elif preferences().space_type=='NODE_EDITOR':
                cat.add("Node")
                cat.add("Options")
    for typename in dir(bpy.types):
        
        try:
            bl_type = getattr(bpy.types, typename,None)
            if issubclass(bl_type, base_type):
                if self.category=='Dropdown':
                    for i,a in enumerate(preferences().dropdown_categories):
                        if i!=self.index:
                            others.extend(split_keep_substring(preferences().dropdown_categories[i].panels))
                    if getattr(bl_type,'backup_space',"None")=='VIEW_3D' or getattr(bl_type,'bl_space_type',"None")=='VIEW_3D':
                        if getattr(bl_type,'bl_category',None) and getattr(bl_type,'bl_category',"None") not in (split_keep_substring(preferences().dropdown_categories[self.index].panels)  if preferences().use_enum_search_for_popups else []):
                            
                            cat.add(getattr(bl_type,'bl_category',"None"))
                        if getattr(bl_type,'backup_category',None) and  getattr(bl_type,'backup_category',"None") not in (split_keep_substring(preferences().dropdown_categories[self.index].panels) if preferences().use_enum_search_for_popups else []):
                            
                            cat.add(getattr(bl_type,'backup_category',"None"))
                elif self.category =='Pie':
                    for i,a in enumerate(preferences().panel_categories):
                        if i!=self.index:
                            others.extend(split_keep_substring(preferences().panel_categories[i].panels))
                    if getattr(bl_type,'backup_space',"None")=='VIEW_3D' or getattr(bl_type,'bl_space_type',"None")=='VIEW_3D':
                        
                        if getattr(bl_type,'bl_category',None) and getattr(bl_type,'bl_category',"None") not in (split_keep_substring(preferences().panel_categories[self.index].panels ) if preferences().use_enum_search_for_popups else []):
                            cat.add(getattr(bl_type,'bl_category',"None"))
                        if getattr(bl_type,'backup_category',None) and getattr(bl_type,'backup_category',"None") not in (split_keep_substring(preferences().panel_categories[self.index].panels) if preferences().use_enum_search_for_popups else []):
                            cat.add(getattr(bl_type,'backup_category',"None"))
                elif self.category =='FP':
                    for i,a in enumerate(preferences().focus_panel_categories):
                        if i!=self.index:
                            others.extend(split_keep_substring(preferences().focus_panel_categories[i].panels))
                    
                    if getattr(bl_type,'backup_space',"None")==preferences().space_type or getattr(bl_type,'bl_space_type',"None")==preferences().space_type:
                        if getattr(bl_type,'bl_category',None) and getattr(bl_type,'bl_category',"None") not in (split_keep_substring(preferences().focus_panel_categories[self.index].panels) if preferences().use_enum_search_for_popups else []):
                            cat.add(getattr(bl_type,'bl_category',"None"))
                        if getattr(bl_type,'renamed_category',None) and getattr(bl_type,'renamed_category',"None") not in (split_keep_substring(preferences().focus_panel_categories[self.index].panels) if preferences().use_enum_search_for_popups else []):
                            cat.add(getattr(bl_type,'renamed_category',"None"))
                        else:
                            if getattr(bl_type,'backup_category',None) and getattr(bl_type,'backup_category',"None") not in (split_keep_substring(preferences().focus_panel_categories[self.index].panels) if preferences().use_enum_search_for_popups else []):
                                cat.add(getattr(bl_type,'backup_category',"None"))
                        
        except:
            pass
    cat=sorted(cat)

    return [(a,a,a,'CHECKMARK' if a in others else 'NONE',i) for i,a in enumerate(cat)]

#addons_to_exclude=['CleanPanels','io_anim_bvh', 'io_curve_svg', 'io_mesh_ply', 'io_mesh_uv_layout', 'io_mesh_stl', 'io_scene_fbx', 'io_scene_gltf2', 'io_scene_obj', 'io_scene_x3d', 'cycles', 'pose_library','node_wrangler', 'node_arrange', 'node_presets','mesh_looptools', 'development_iskeyfree','development_icon_get','add_curve_extra_objects', 'add_mesh_extra_objects','space_view3d_spacebar_menu', 'development_edit_operator']
# addons_to_exclude=['CleanPanels','io_anim_bvh', 'io_curve_svg', 'io_mesh_ply', 'io_mesh_uv_layout', 'io_mesh_stl', 'io_scene_fbx', 'io_scene_gltf2', 'io_scene_obj', 'io_scene_x3d', 'cycles', 'node_presets', 'development_iskeyfree','development_icon_get','add_curve_extra_objects', 'add_mesh_extra_objects','space_view3d_spacebar_menu', 'development_edit_operator', 'add_camera_rigs','add_curve_sapling', 'add_mesh_BoltFactory', 'add_mesh_discombobulator', 'add_mesh_geodesic_domes', 'blender_id,btrace', 'system_blend_info', 'system_property_chart', 'io_anim_camera', 'io_export_dxf', 'io_export_pc2', 'io_import_BrushSet', 'io_import_dxf', 'io_import_images_as_planes', 'mesh_bsurfaces', 'context_browser', 'io_mesh_atomic', 'io_import_palette', 'io_scene_usdz', 'io_shape_mdd', 'lighting_tri_lights', 'lighting_dynamic_sky', 'mesh_inset', 'ui_translate', 'clouds_generator', 'blender_id', 'btrace', 'curve_assign_shapekey', 'curve_simplify', 'depsgraph_debug', 'sun_position', 'mesh_auto_mirror', 'mesh_f2', 'mesh_snap_utilities_line', 'MSPlugin', 'NodePreview', 'object_fracture_cell', 'object_scatter', 'object_skinify', 'render_copy_settings', 'space_view3d_3d_navigation', 'space_view3d_brush_menus', 'space_view3d_copy_attributes', 'space_view3d_math_vis', 'object_carver', 'object_color_rules', 'render_freestyle_svg', 'render_ui_animation_render', 'space_view3d_modifier_tools', 'space_view3d_pie_menus']
addons_to_exclude=['CleanPanels']
DIRNAME=os.path.dirname(__file__)
if os.path.isfile(os.path.join(DIRNAME,"Addons to exclude.txt")):
    with open(os.path.join(DIRNAME,"Addons to exclude.txt"),newline='\n',encoding='utf-8') as f:
        for a in f.readlines():
            addons_to_exclude.append(a.strip())
# print(addons_to_exclude)
def get_all_python_files(dir_path):
    pyfiles = []
    for path, subdirs, files in os.walk(dir_path):
        for name in files:
            if name.endswith('.py') and os.path.join(path, name)!=__file__:
                pyfiles.append(os.path.join(path, name))
                #print(os.path.join(path, name))
    return pyfiles
import shutil
def get_icons(self, context):
    icons=bpy.types.UILayout.bl_rna.functions[
                "prop"].parameters["icon"].enum_items.keys()
    return [(a,a,a,a,i) for i,a in enumerate(icons)] 
ALL_ICONS_ENUM=get_icons(None,None)
ALL_ICONS=[a for a,_,_,_,_ in get_icons(None,None)]
def clean_all_python_files(remove=False):
    #directory=os.path.dirname(os.path.dirname(__file__))
    directories=[]
    addons=[]
    a=preferences().addon_info[preferences().addon_info_index]
    if remove:
        if a.addons:
            addons.extend(split_keep_substring(a.addons))
    else:
        if a.ordered:
            addons.extend(split_keep_substring(a.ordered))
    #print(addons)
    for mod in addon_utils.modules():
        if mod.__name__ in addons:
            directories.append(os.path.dirname(mod.__file__) if os.path.dirname(mod.__file__)!=os.path.dirname(os.path.dirname(__file__)) else mod.__file__)
            
        else:
            pass
    if not os.path.isdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels")):
        os.mkdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels"))
    for directory in directories:
        if os.path.isfile(directory):
            if not remove and preferences().auto_backup_addons:
                shutil.copy(directory, os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels",os.path.splitext(os.path.basename(directory))[0]+str(datetime.now().strftime(r' %d-%m-%Y %H %M'))+os.path.splitext(os.path.basename(directory))[1]))
                print("Creating Backup....\n", os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels",os.path.splitext(os.path.basename(directory))[0]+str(datetime.now().strftime(r' %d-%m-%Y %H %M'))+os.path.splitext(os.path.basename(directory))[1]))
            f=directory
            #print(f)
            try:
                data=[]
                with open(f,mode='r') as file:
                    while True:
                        line=file.readline()
                        if not line:
                            break
                        if remove:
                            data.append(line.replace("#--editied-by-CleanPanels--",""))
                        else:
                            data.append(line.replace("bl_order","#--editied-by-CleanPanels--bl_order"))
                if data:
                    with open(f,mode='w') as file:
                            file.writelines(data)
            except Exception as e:
                log("Error in file",f,'\n',e)
        else:
            #print(directory,os.path.dirname(os.path.dirname(__file__)))
            if directory!=os.path.dirname(os.path.dirname(__file__)):
                if not remove and preferences().auto_backup_addons:
                    dest_path=os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))),"Addon-Backups-CleanPanels",os.path.basename(directory)+str(datetime.now().strftime(r' %d-%m-%Y %H %M')))
                    if not os.path.isdir(dest_path):
                        shutil.copytree(directory, dest_path)
                        print("Creating Backup....\n", dest_path)
                for f in get_all_python_files(directory):
                    #print(f)
                    try:
                        data=[]
                        with open(f,mode='r') as file:
                            while True:
                                line=file.readline()
                                if not line:
                                    break
                                if remove:
                                    data.append(line.replace("#--editied-by-CleanPanels--",""))
                                else:
                                    data.append(line.replace("bl_order","#--editied-by-CleanPanels--bl_order"))
                        if data:
                            with open(f,mode='w') as file:
                                    file.writelines(data)
                    except Exception as e:
                        log("Error in file",f,'\n',e)
import time
import numpy as np
def sort_panels2(panels,order):
    # st=time.time()
    sorted=[]
    used=[]
    bl_types=[a[1] for a in panels]
    order=[a.replace("\n","") for a in order if a in bl_types]
    for o in order:
        for b,p in panels:
            if p.__name__==o and p not in used:
                used.append(p)
                sorted.append((b,p))
                break
    for b,p in panels:
        if p not in used:
            sorted.append((b,p))
    # print(time.time()-st)        
    return sorted
def sort_by_another_list(source_list,ordered_list):
    keys = { ordered_list[i]: i for i in range(len(ordered_list)) }
    sorted_list = sorted(source_list, key=lambda x: keys[x] if x in keys else len(source_list)+1)   
    return sorted_list
def sort_panels(panels,order):
    order=view_panels+tool_panels+order
    keys = { order[i]: i for i in range(len(order)) }
    sorted_panels = sorted(panels, key=lambda x: keys[x[1].__name__] if x[1].__name__ in keys else len(panels)+1)  
    return sorted_panels
def sort_panels_dropdowns(panels,order):
    keys = { order[i]: i for i in range(len(order)) }
    sorted_panels = sorted(panels, key=lambda x: keys[x[0].__name__] if x[0].__name__ in keys else len(panels)+1)  
    return sorted_panels
def register_panel(c):
    try: 
        bpy.utils.register_class(c,False)
    except:
        bpy.utils.register_class(c)
def draw_dropdowns(self, context):
    layout_str=str(self.layout.introspect())
    already_drawn='bpy.ops.cp.popup' in layout_str
    if not already_drawn:
        if preferences().use_dropdowns:
            categories=[]
            layout=self.layout.row()
            layout.separator()
            if preferences().show_dropdown_search:
                layout.operator("cp.search_dropdown",icon='VIEWZOOM',text='Search')
            for a in preferences().dropdown_categories:
                    if a.name==context.scene.pap_active_dropdown_category:
                        categories=split_keep_substring(a.panels)
                        categories=[a.strip() for a in categories]
            for a in categories:
                if a:

                    layout.emboss='PULLDOWN_MENU'
                    layout.operator("cp.popupcompletepanel",text=a,icon="DOWNARROW_HLT").name=a
def panel_opened(self, context):
    if context.scene.pap_opened_panels.find(self.name)>=0:
        t=context.scene.pap_opened_panels[context.scene.pap_opened_panels.find(self.name)]
    else:
        t=context.scene.pap_opened_panels.add()
        t.name=self.name
    t.opened_before=True
    t.pap_opened_panels=""
    for i in range(1,39):
        if getattr(self,f"show_panel_{i}",False):
            t.pap_opened_panels=t.pap_opened_panels+","+str(i) if t.pap_opened_panels else str(i)

def get_current_context(context):
    if context.mode=='OBJECT':
        return "objectmode"
    elif context.mode =='EDIT_MESH':
        return "mesh_edit"
    elif context.mode =='EDIT_CURVE':
        return "curve_edit"
    elif context.mode =='EDIT_SURFACE':
        return "surface_edit"
    elif context.mode =='EDIT_TEXT':
        return "text_edit"
    elif context.mode =='SCULPT':
        return "sculpt_mode"
    elif context.mode =='EDIT_ARMATURE':
        return "armature_edit"
    elif context.mode =='EDIT_METABALL':
        return "mball_edit"
    elif context.mode =='EDIT_LATTICE':
        return "lattice_edit"
    elif context.mode =='POSE':
        return "posemode"
    elif context.mode =='PAINT_WEIGHT':
        return "weightpaint"
    elif context.mode =='PAINT_VERTEX':
        return "vertexpaint"
    elif context.mode =='PAINT_TEXTURE':
        return "imagepaint"
    elif context.mode =='PARTICLE':
        return "particlemode"
    else:
        return "None1"
def check_for_injection():
    version=bpy.app.version
    if sys.platform=='darwin':
        util_file_path=os.path.join(os.path.dirname(os.path.dirname(bpy.app.binary_path)),'Resources',f"{version[0]}.{version[1]}","scripts","modules","addon_utils.py")
    else:
        util_file_path=os.path.join(os.path.dirname(bpy.app.binary_path),f"{version[0]}.{version[1]}","scripts","modules","addon_utils.py")
    if os.path.isfile(util_file_path):
        with open(util_file_path,mode='r') as f:
            text=f.read()
            path=os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config", "CP-config.txt")
            if path in text:
                return True
    return False
def check_for_tracking_injection():
    version=bpy.app.version
    if sys.platform=='darwin':
        util_file_path=os.path.join(os.path.dirname(os.path.dirname(bpy.app.binary_path)),'Resources',f"{version[0]}.{version[1]}","scripts","modules","bpy","utils","__init__.py")
    else:
        util_file_path=os.path.join(os.path.dirname(bpy.app.binary_path),f"{version[0]}.{version[1]}","scripts","modules","bpy","utils","__init__.py")
    if os.path.isfile(util_file_path):
        with open(util_file_path,mode='r') as f:
            text=f.read()
            path=os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config", "CP-PanelOrder.txt")
            if path in text:    
                return True
    return False
def check_for_delayed_loading_injection():
    version=bpy.app.version
    if sys.platform=='darwin':
        util_file_path=os.path.join(os.path.dirname(os.path.dirname(bpy.app.binary_path)),'Resources',f"{version[0]}.{version[1]}","scripts","modules","addon_utils.py")
    else:
        util_file_path=os.path.join(os.path.dirname(bpy.app.binary_path),f"{version[0]}.{version[1]}","scripts","modules","addon_utils.py")
    if os.path.isfile(util_file_path):
        with open(util_file_path,mode='r') as f:
            text=f.read()
            path=os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config", "CP-Addons To Load on Boot.txt")
            
            if f"atl_file=r'{path}'" in text:   
                return True
    return False
def get_module_name(bl_type):
    package_name=inspect.getmodule(bl_type).__name__
    if "." in package_name:
        name=package_name[:package_name.index(".")]
    else:
        name=package_name
    if name in exceptional_names.keys():
        name=exceptional_names[name]
    name=get_custom_module_name(bl_type,name)
    return name

def change_panel_category(old_name,new_name,space='VIEW_3D'):
    registered_panels=[]
    for typename in dir(bpy.types):
        
        try:
            bl_type = getattr(bpy.types, typename,None)
            if issubclass(bl_type, bpy.types.Panel):
                package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                if package_name=='__main__' or package_name=='Brush_Manager':
                    continue
                
                registered_panels.append(bl_type)
                if not getattr(bl_type,'backup_category',None) and getattr(bl_type,'bl_category'):
                    bl_type.backup_category=bl_type.bl_category
        except:
            pass
    config_path=os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config","CP-PanelOrder.txt")
    order_of_panels=[]
    if os.path.isfile(config_path):
        with open(config_path, mode='r', newline='\n', encoding='utf-8') as file:
            order_of_panels=file.readlines()
    else:
        if os.path.isfile(os.path.join(os.path.dirname(__file__),'CP-PanelOrder.txt')):
            with open(os.path.join(os.path.dirname(__file__),'CP-PanelOrder.txt'), mode='r', newline='\n', encoding='utf-8') as file:
                order_of_panels=file.readlines()
    cleaned_order=[]

    if hasattr(bpy.utils,'panels'):
        if len(getattr(bpy.utils,'panels',[]))>len(order_of_panels):
            order_of_panels=getattr(bpy.utils,'panels',[])
    for o in order_of_panels:
        if o.replace("\n","") not in cleaned_order:
            cleaned_order.append(o.replace("\n",""))
    order_of_panels=cleaned_order
    categories=[]
    for a in split_keep_substring(preferences().addons_to_exclude)+addons_to_exclude:
        try:
            if get_module_name_from_addon_name(a)!='--Unknown--':
                categories.append(get_module_name_from_addon_name(a))
        except:
            pass
    for index,a in enumerate(preferences().workspace_categories):
        if getattr(preferences().categories,f'enabled_{index}',False):
            #if a.name==context.workspace.pap_active_workspace_category:
                #categories_string= ''.join(a.panels.split())
                categories_string=split_keep_substring(a.panels)
                #categories.extend([a.strip() for a in categories_string])
                categories.extend([get_module_name_from_addon_name(a) for a in categories_string])
    panels_to_reregister=[]
    parents=[]
    children=[]
    parents_to_move_back=[]
    children_to_move_back=[]
    panels_to_move_back=[]
    for bl_type in registered_panels:
        try:
            package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
            if "." in package_name:
                name=package_name[:package_name.index(".")]
            else:
                name=package_name
            if name in exceptional_names.keys():
                name=exceptional_names[name]
            name=get_custom_module_name(bl_type,name)
            try:
                is_panel=False
                try:
                    is_panel=issubclass(bl_type, bpy.types.Panel)
                except:
                    pass                    
                if bl_type and is_panel:
                    if (getattr(bl_type,'bl_category',None) and getattr(bl_type,'backup_space',None)==space and getattr(bl_type,"backup_region","None")=='UI' ) and not getattr(bl_type,'bl_parent_id',None):
                        package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                        if "." in package_name:
                            name=package_name[:package_name.index(".")]
                        else:
                            name=package_name
                        if name in exceptional_names.keys():
                            name=exceptional_names[name]
                        name=get_custom_module_name(bl_type,name)
                        if name!="bl_ui":
                            
                            if getattr(bl_type,'bl_category',None)==old_name or(getattr(bl_type,'bl_category',None)==preferences().holder_tab_name and getattr(bl_type,'renamed_category',None)==old_name):
                                #bl_type.bl_category=bl_type.backup_category if getattr(bl_type,'backup_category',None) else bl_type.bl_category
                                
                                #print(bl_type,getattr(bl_type,'backup_category',None))
                                panels_to_reregister.append((name,bl_type))
                                
                            else:
                                if  getattr(bl_type,'bl_category',None)=='Focused':
                                    panels_to_move_back.append((name,bl_type))
                            # bpy.utils.unregister_class(bl_type)  
                            # bpy.utils.register_class(bl_type)
                    else:
                        if bl_type and getattr(bl_type,'bl_parent_id',None):
                            package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                            if "." in package_name:
                                name=package_name[:package_name.index(".")]
                            else:
                                name=package_name
                            name=get_custom_module_name(bl_type,name)
                            if name in exceptional_names.keys():
                                        name=exceptional_names[name]
                            if name!="bl_ui":
                                if getattr(bl_type,'bl_category',None)==old_name or(getattr(bl_type,'renamed_category',None)==old_name):
                                    pass
                                    #bl_type.bl_category=bl_type.backup_category if getattr(bl_type,'backup_category',None) else bl_type.bl_category
                                    #print(bl_type,getattr(bl_type,'backup_category',None))
                                    parents.append(bl_type.bl_parent_id)
                                    children.append((name,bl_type))
                                    # bpy.utils.unregister_class(bl_type)  
                                    # bpy.utils.register_class(bl_type)
                                else:
                                    if getattr(bl_type,'bl_category',None)=='Focused' :
                                        parents_to_move_back.append(bl_type.bl_parent_id)
                                        children_to_move_back.append((name,bl_type))
            except Exception as e:
                #pass
                print(e)
        except:
            pass
    panels_to_reregister=sort_panels(panels_to_reregister,order_of_panels)
    children=sort_panels(children,order_of_panels)
    children=sorted(children,key=lambda x:getattr(x[1],'bl_parent_id','Temp') in [getattr(a[1],'bl_idname','None') for a in children])
    # print(panels_to_reregister)
    # print("")
    # print("")
    # print("")
    # print("")
    # print(children,"Children")
    # for c in children:
    #      print(getattr(c,'bl_category',None))
    # print("")
    # print("")
    
    # print(panels_to_move_back)
    for name,p in panels_to_reregister:
        if getattr(p,'backup_space',None)!=None and getattr(p,'backup_region',None)!=None:
            p.bl_space_type=p.backup_space
            p.bl_region_type=p.backup_region
        if not getattr(p,'backup_category',None):
            p.backup_category=p.bl_category
        if getattr(p,'backup_order',None)!=None:
                                p.bl_order=getattr(p,'backup_order')
        p.bl_category=new_name

        try:
            bpy.utils.unregister_class(p)  
            register_panel(p)
        except Exception as e:
            print(e)
    for name,c in children:
        if getattr(c,'backup_space',None)!=None and getattr(c,'backup_region',None)!=None:
            c.bl_space_type=c.backup_space
            c.bl_region_type=c.backup_region
        if getattr(c,'backup_order',None)!=None:
                                c.bl_order=getattr(c,'backup_order')
        if getattr(c,'bl_category',None):
            if not getattr(c,'backup_category',None):
                c.backup_category=c.bl_category
        c.bl_category=new_name
        try:
            if c.bl_parent_id in parents:
                bpy.utils.unregister_class(c)  
                register_panel(c)
        except Exception as e:
                    print(e)
    for name,c in children:
        if getattr(c,'backup_space',None)!=None and getattr(c,'backup_region',None)!=None:
            c.bl_space_type=c.backup_space
            c.bl_region_type=c.backup_region
        if getattr(c,'backup_order',None)!=None:
                                c.bl_order=getattr(c,'backup_order')
        if getattr(c,'bl_category',None):
            if not getattr(c,'backup_category',None):
                c.backup_category=c.bl_category
        c.bl_category=new_name
        try:
            if c.bl_parent_id not in parents:
                bpy.utils.unregister_class(c)  
                register_panel(c)
        except Exception as e:
                    print(e)
    for name,p in panels_to_move_back:
        if name in categories or name.replace("ender-","") in categories:
            if preferences().addon_info_for_renaming.find(name)>=0:
                if getattr(p,'bl_category'):
                    if not getattr(p,'backup_category',None):
                        p.backup_category=p.bl_category
                if getattr(p,'backup_order',None)!=None:
                                p.bl_order=getattr(p,'backup_order')
                p.bl_category=preferences().addon_info_for_renaming[preferences().addon_info_for_renaming.find(name)].tab_name
                p.renamed_category=p.bl_category
            else:
                if getattr(p,'backup_category',None):   
                    p.bl_category=getattr(p,'backup_category',None)
        else:
            p.bl_category=preferences().holder_tab_name
            if getattr(p,'backup_order',None)!=None:
                                p.bl_order=getattr(p,'backup_order')
        try:
            bpy.utils.unregister_class(p)  
            register_panel(p)
        except Exception as e:
            print(e)
    for name,c in children_to_move_back:
                if preferences().addon_info_for_renaming.find(name)>=0:
                    if getattr(c,'bl_category',None):
                        if not getattr(c,'backup_category',None):
                            c.backup_category=c.bl_category
                    if getattr(c,'backup_order',None)!=None:
                                c.bl_order=getattr(c,'backup_order')
                    c.bl_category=preferences().addon_info_for_renaming[preferences().addon_info_for_renaming.find(name)].tab_name
                    c.renamed_category=c.bl_category
                else:
                    if getattr(c,'backup_category',None):   
                        c.bl_category=getattr(c,'backup_category',None)
                try:
                    if c.bl_parent_id in parents_to_move_back:
                        bpy.utils.unregister_class(c)  
                        register_panel(c)
                except Exception as e:
                            print(e)

    for name,c in children_to_move_back:
        if preferences().addon_info_for_renaming.find(name)>=0:
            if getattr(c,'bl_category',None):
                if not getattr(c,'backup_category',None):
                    c.backup_category=c.bl_category
            if getattr(c,'backup_order',None)!=None:
                                c.bl_order=getattr(c,'backup_order')
            c.bl_category=preferences().addon_info_for_renaming[preferences().addon_info_for_renaming.find(name)].tab_name
            c.renamed_category=c.bl_category
        else:
                if getattr(c,'backup_category',None):   
                    c.bl_category=getattr(c,'backup_category',None)
        try:
            if c.bl_parent_id not in parents_to_move_back:
                bpy.utils.unregister_class(c)  
                register_panel(c)
        except Exception as e:
                    print(e)
icon_collection = {}
def load_icons(only_custom=False):
    
    global icon_collection
    if "icons" in icon_collection.keys():
        pcoll=icon_collection["icons"]
    else:
        pcoll = bpy.utils.previews.new()
    loaded_icons=[]
    if not only_custom:
        my_icons_dir = os.path.join(os.path.dirname(__file__), "icons")
        for file in os.listdir(my_icons_dir):
            icon_name=os.path.splitext(file)[0].lower()
            if icon_name not in pcoll.keys():
                pcoll.load(icon_name, os.path.join(my_icons_dir, file), 'IMAGE')
    custom_icons_dir=preferences().custom_icons_dir
    if os.path.isdir(custom_icons_dir):
        for file in os.listdir(custom_icons_dir):
            if file.endswith('.png'):
                icon_name=os.path.splitext(file)[0].lower()
                if icon_name not in pcoll.keys():
                    pcoll.load(icon_name, os.path.join(custom_icons_dir, file), 'IMAGE')
    # pcoll.load("youtube", os.path.join(my_icons_dir, "Youtube.png"), 'IMAGE')
    # pcoll.load("discord", os.path.join(my_icons_dir, "Discord.png"), 'IMAGE')
    # pcoll.load("updatered", os.path.join(my_icons_dir, "updatered.png"), 'IMAGE')
    # pcoll.load("updategreen", os.path.join(my_icons_dir, "updategreen.png"), 'IMAGE')
    # pcoll.load("city", os.path.join(my_icons_dir, "City.png"), 'IMAGE')
    # pcoll.load("landscape", os.path.join(my_icons_dir, "landscape.png"), 'IMAGE')
    # pcoll.load("animal", os.path.join(my_icons_dir, "Animal.png"), 'IMAGE')
    # pcoll.load("tree", os.path.join(my_icons_dir, "Tree.png"), 'IMAGE')
    # pcoll.load("car", os.path.join(my_icons_dir, "Car.png"), 'IMAGE')
    # pcoll.load("transport", os.path.join(my_icons_dir, "Transport.png"), 'IMAGE')
    # pcoll.load("truck", os.path.join(my_icons_dir, "Truck.png"), 'IMAGE')
    # pcoll.load("reorder", os.path.join(my_icons_dir, "Reorder.png"), 'IMAGE')
    icon_collection["icons"] = pcoll
prefs_to_save = {'use_enum_search_for_popups':'bool',
     'filter_internal_tabs':'bool',
     'favorites':'pc',
                 'panel_categories':'pc',
                 'pop_out_style':'str',
                 'dropdown_categories':'pc',
                 'workspace_categories':'pc',
                 'focus_panel_categories':'pc',
                 'addons_to_exclude':'str',
                 
                 'addon_info':'order',
                 'addon_info_for_renaming':'order',

                 'workspace_categories_image_editor':'pc',
                 'focus_panel_categories_image_editor':'pc',
                 'addons_to_exclude_image_editor':'str',
                 
                 'addon_info_image_editor':'order',
                 'addon_info_for_renaming_image_editor':'order',

                 'workspace_categories_node_editor':'pc',
                 'focus_panel_categories_node_editor':'pc',
                 'addons_to_exclude_node_editor':'str',
                 
                 'addon_info_node_editor':'order',
                 'addon_info_for_renaming_node_editor':'order',

                 'draw_side':'str',
                 'addon_desc_info':'order',
                 'experimental':'bool',
                 'use_sticky_popup':'bool',
                 'columm_layout_for_popup':'str',
                 'use_verticle_menu':'bool',
                 'dropdown_width':'int',
                 'show_dropdown_search':'bool',
                 'auto_backup_addons':'bool',
                 'filtering_method':'str',
                 'show_advanced':'bool',
                 'use_dropdowns':'bool',
                 'sort_per_category':'bool',
                 'holder_tab_name':'str',
                 'custom_icons_dir':'str',
                 'sort_focus_menu_based_on_clicks':'bool',
                 'only_show_unfiltered_panels':'bool',
                 'filtering_per_workspace':'bool',
                 'show_delete_buttons_in_quick_settings':'bool',
                 'move_dropdowns_to_toolbar':'bool'

}
def savePreferences(self=None,context=None):
    savePreferencesToPath()
    
def read_time_from_file(filename):
    with open(filename, 'r') as file:
        return file.read().strip()
def save_time_to_file(filename, time_str):
    with open(filename, 'w+') as file:
        file.write(time_str)
def create_backup_configs():
    if len(preferences().panel_categories)>0 or len(preferences().dropdown_categories)>0 or len(preferences().focus_panel_categories)>0 or len(preferences().workspace_categories)>0:
        current_time_iso=datetime.now().isoformat()
        if os.path.exists(os.path.join(os.path.dirname(__file__),"last_backup.txt")):
            try:
                last_time=read_time_from_file(os.path.join(os.path.dirname(__file__),"last_backup.txt"))
                current_time_dt = datetime.fromisoformat(current_time_iso)
                time_from_file_dt = datetime.fromisoformat(last_time)
                if abs((current_time_dt-time_from_file_dt).days)<7:
                    return
            except:
                pass
        try:
            save_time_to_file(os.path.join(os.path.dirname(__file__),"last_backup.txt"),current_time_iso)
        except Exception:
            pass
        savePreferencesToPath(True)
def savePreferencesToPath(backup=False):
    config_folder_path=os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config")
    config_file_path=os.path.join(config_folder_path, "CP-config.txt")
    atl_file_path=os.path.join(config_folder_path, "CP-Addons To Load on Boot.txt")
    if backup:
         config_folder_path=os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config","CP-Backups")
         config_file_path=os.path.join(config_folder_path, f"CP-config-{str(datetime.now().strftime(r' %d-%m-%Y %H %M'))}.txt")
    if not os.path.isdir(config_folder_path):
        os.mkdir(config_folder_path)
    with open(atl_file_path, mode='w+', newline='\n', encoding='utf-8') as file:
        for line in preferences().atl_list.split(","):
            file.write(f"{line}\n")
    with open(config_file_path, mode='w+', newline='\n', encoding='utf-8') as file:
        for p, t in prefs_to_save.items():
            if p == 'favorites':
                for s in preferences().favorites:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.panels}\n")
            if p == 'panel_categories':
                for s in preferences().panel_categories:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.panels}\n")
            elif p == 'dropdown_categories':
                for s in preferences().dropdown_categories:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.panels}\n")
            elif p == 'focus_panel_categories':
                for s in preferences().focus_panel_categories:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.panels}\n")
            elif p == 'workspace_categories':
                for s in preferences().workspace_categories:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.panels}===Icon=>{s.icon}\n")
            elif p == 'addon_info':
                for s in preferences().addon_info:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.addons}===Ordered=>{s.ordered}\n")
            elif p == 'addon_info_for_renaming':
                for s in preferences().addon_info_for_renaming:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.tab_name}\n")
            elif p == 'focus_panel_categories_node_editor':
                for s in preferences().focus_panel_categories_node_editor:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.panels}\n")
            elif p == 'workspace_categories_node_editor':
                for s in preferences().workspace_categories_node_editor:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.panels}===Icon=>{s.icon}\n")
            elif p == 'addon_info_node_editor':
                for s in preferences().addon_info_node_editor:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.addons}===Ordered=>{s.ordered}\n")
            elif p == 'addon_info_for_renaming_node_editor':
                for s in preferences().addon_info_for_renaming_node_editor:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.tab_name}\n")
            elif p == 'focus_panel_categories_image_editor':
                for s in preferences().focus_panel_categories_image_editor:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.panels}\n")
            elif p == 'workspace_categories_image_editor':
                for s in preferences().workspace_categories_image_editor:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.panels}===Icon=>{s.icon}\n")
            elif p == 'addon_info_image_editor':
                for s in preferences().addon_info_image_editor:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.addons}===Ordered=>{s.ordered}\n")
            elif p == 'addon_info_for_renaming_image_editor':
                for s in preferences().addon_info_for_renaming_image_editor:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.tab_name}\n")
            elif p == 'addon_desc_info':
                for s in preferences().addon_desc_info:
                    file.write(
                        f"{p}=>{t}==={s.name}>>{s.desc}\n")
            else:
                file.write(f"{p}=>{t}==={getattr(preferences(),p)}\n")
        order=[]
        for a in preferences().addon_info:
            order.extend(split_keep_substring(a.addons))
        #print(order)
        file.write("addon_order=>order===Category>>"+",".join(order)+"\n")

@persistent
def loadPreferences(a=None,b=None):
    
    all_icons=ALL_ICONS+list(icon_collection["icons"].keys())
    if not os.path.isdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config")):
        os.mkdir(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(
            os.path.dirname(os.path.abspath(__file__))))), "config"))
    if os.path.isfile(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config", "CP-config.txt")):
        with open(os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config", "CP-config.txt"), mode='r', newline='\n', encoding='utf-8') as file:
            prefs = file.readlines()
            reset_preferences()
            for p in prefs:
                try:
                    attr = p[:p.index("=>")]
                    type = p[p.index("=>")+2:p.index("===")]
                    value = p[p.index("===")+3:]
                    value = value.replace("\n", "")
                    if attr =='favorites' and type=='pc':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        panels=value[value.index(">>")+2:]
                        pc=preferences().favorites.add()
                        pc.name=name
                        pc.panels=panels
                    if attr =='panel_categories' and type=='pc':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        panels=value[value.index(">>")+2:]
                        pc=preferences().panel_categories.add()
                        pc.name=name
                        pc.panels=panels
                    elif attr =='dropdown_categories' and type=='pc':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        panels=value[value.index(">>")+2:]
                        pc=preferences().dropdown_categories.add()
                        pc.name=name
                        pc.panels=panels
                    elif attr =='focus_panel_categories' and type=='pc':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        panels=value[value.index(">>")+2:]
                        pc=preferences().focus_panel_categories.add()
                        pc.name=name
                        pc.panels=panels
                    elif attr =='workspace_categories' and type=='pc':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        panels=value[value.index(">>")+2:value.index("Icon=>")-3]
                        icon=value[value.index("Icon=>")+6:]
                        pc=preferences().workspace_categories.add()
                        pc.name=name
                        pc.panels=panels
                        pc.icon=icon if icon in all_icons else 'NONE'
                    elif attr =='addon_info' and type=='order':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        addons=value[value.index(">>")+2:value.index("Ordered=>")-3]
                        ordered=value[value.index("Ordered=>")+9:]
                        pc=preferences().addon_info.add()
                        pc.name=name
                        pc.addons=addons
                        pc.ordered=ordered
                    elif attr =='addon_info_for_renaming' and type=='order':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        tab_name=value[value.index(">>")+2:]
                        pc=preferences().addon_info_for_renaming.add()
                        pc.name=name
                        pc.tab_name=tab_name
                    elif attr =='focus_panel_categories_image_editor' and type=='pc':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        panels=value[value.index(">>")+2:]
                        pc=preferences().focus_panel_categories_image_editor.add()
                        pc.name=name
                        pc.panels=panels
                    elif attr =='workspace_categories_image_editor' and type=='pc':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        panels=value[value.index(">>")+2:value.index("Icon=>")-3]
                        icon=value[value.index("Icon=>")+6:]
                        pc=preferences().workspace_categories_image_editor.add()
                        pc.name=name
                        pc.panels=panels
                        pc.icon=icon if icon in all_icons else 'NONE'
                    elif attr =='addon_info_image_editor' and type=='order':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        addons=value[value.index(">>")+2:value.index("Ordered=>")-3]
                        ordered=value[value.index("Ordered=>")+9:]
                        pc=preferences().addon_info_image_editor.add()
                        pc.name=name
                        pc.addons=addons
                        pc.ordered=ordered
                    elif attr =='addon_info_for_renaming_image_editor' and type=='order':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        tab_name=value[value.index(">>")+2:]
                        pc=preferences().addon_info_for_renaming_image_editor.add()
                        pc.name=name
                        pc.tab_name=tab_name
                    elif attr =='focus_panel_categories_node_editor' and type=='pc':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        panels=value[value.index(">>")+2:]
                        pc=preferences().focus_panel_categories_node_editor.add()
                        pc.name=name
                        pc.panels=panels
                    elif attr =='workspace_categories_node_editor' and type=='pc':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        panels=value[value.index(">>")+2:value.index("Icon=>")-3]
                        icon=value[value.index("Icon=>")+6:]
                        pc=preferences().workspace_categories_node_editor.add()
                        pc.name=name
                        pc.panels=panels
                        pc.icon=icon if icon in all_icons else 'NONE'
                    elif attr =='addon_info_node_editor' and type=='order':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        addons=value[value.index(">>")+2:value.index("Ordered=>")-3]
                        ordered=value[value.index("Ordered=>")+9:]
                        pc=preferences().addon_info_node_editor.add()
                        pc.name=name
                        pc.addons=addons
                        pc.ordered=ordered
                    elif attr =='addon_info_for_renaming_node_editor' and type=='order':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        tab_name=value[value.index(">>")+2:]
                        pc=preferences().addon_info_for_renaming_node_editor.add()
                        pc.name=name
                        pc.tab_name=tab_name
                    elif attr =='addon_desc_info':
                        
                        # value=value.replace("[","").replace("]","")
                        name=value[:value.index(">>")]
                        desc=value[value.index(">>")+2:]
                        pc=preferences().addon_desc_info.add()
                        pc.name=name
                        pc.desc=desc
                    elif type=='bool' or type=='int':
                        setattr(preferences(), attr, eval(value))
                    
                    else:
                        
                        setattr(preferences(), attr, value)
                    
                except Exception as e:
                    pass
        setattr(preferences(), 'injected_code', check_for_injection())
        setattr(preferences(), 'injected_code_tracking', check_for_tracking_injection())
        setattr(preferences(), 'delayed_loading_code_injected', check_for_delayed_loading_injection())
        setattr(preferences(),'delayed_addons_loaded',False)

def reset_preferences():
    preferences().favorites.clear()
    preferences().panel_categories.clear()
    preferences().dropdown_categories.clear()
    preferences().focus_panel_categories.clear()
    preferences().workspace_categories.clear()
    preferences().addon_info.clear()
    preferences().addon_info_for_renaming.clear()
    preferences().focus_panel_categories_node_editor.clear()
    preferences().workspace_categories_node_editor.clear()
    preferences().addon_info_node_editor.clear()
    preferences().addon_info_for_renaming_node_editor.clear()
    preferences().focus_panel_categories_image_editor.clear()
    preferences().workspace_categories_image_editor.clear()
    preferences().addon_info_image_editor.clear()
    preferences().addon_info_for_renaming_image_editor.clear()
    preferences().addon_desc_info.clear()
    preferences().addons_to_exclude=""
    preferences().addons_to_exclude_image_editor=""
    preferences().addons_to_exclude_node_editor=""
def remove_duplicates(list):
    result=[]
    for a in list:
        if a not in result:
            result.append(a)
    return result
def workspace_category_enabled(self, context):
    try:
        space_type=context.area.type
    except:
        space_type='VIEW_3D'
    try:
        if context.area.type=='PREFERENCES':
            space_type=preferences().space_type
    except:
        space_type='VIEW_3D'
    
    if not preferences().filtering_method=="Use N-Panel Filtering":
        if self.filter_enabled:
            context.workspace.use_filter_by_owner = True
            categories=[]
            
            for index,a in enumerate(getattr(preferences(),f"workspace_categories{get_active_space(space_type)}")):
                if getattr(self,f'enabled{get_active_space(space_type)}_{index}',False):
                    #if a.name==context.workspace.pap_active_workspace_category:
                        #categories_string= ''.join(a.panels.split())
                        categories_string=split_keep_substring(a.panels)
                        #categories.extend([a.strip() for a in categories_string])
                        categories.extend([get_module_name_from_addon_name(a) for a in categories_string])
            
            for a in [__package__] + categories[:]:
                try:
                    a=sys.modules[a].__name__
                    if a not in [c.name for c in context.workspace.owner_ids]:
                        bpy.ops.wm.owner_enable(owner_id=a)
                except:
                    pass
            for a in split_keep_substring(getattr(preferences(),f"addons_to_exclude{get_active_space(space_type)}"))+addons_to_exclude:
                a=get_module_name_from_addon_name(a)
                try:
                    if a not in [c.name for c in context.workspace.owner_ids] and a in bpy.context.preferences.addons.keys():
                        bpy.ops.wm.owner_enable(owner_id=a)
                except:
                    pass
            for b in bpy.context.preferences.addons.keys():
                try:
                    #print(b)
                    mod = sys.modules[b]
                    if mod.__name__ not in categories+[__package__] and mod.__name__ in [a.name for a in context.workspace.owner_ids]:
                        if mod.__name__ not in [get_module_name_from_addon_name(a) for a in split_keep_substring(getattr(preferences(),f"addons_to_exclude{get_active_space(space_type)}"))]+addons_to_exclude:
                            #print("Disable",mod.__name__)
                            bpy.ops.wm.owner_disable(owner_id=mod.__name__)
                except:
                        pass
        else:
            context.workspace.use_filter_by_owner = False
    else:
        registered_panels=[]
        for typename in dir(bpy.types):
            try:
                bl_type = getattr(bpy.types, typename,None)
                if issubclass(bl_type, bpy.types.Panel):
                    # if getattr(bl_type,"bl_region_type","None")=='UI' and getattr(bl_type,'bl_space_type',None)==space_type:
                    #     registered_panels.append(bl_type)
                    

                    package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                    if package_name=='__main__':
                        continue
                    # if "mega" or "liq" in package_name:
                    #      print(bl_type,getattr(bl_type,'bl_category','None'))
                    if "." in package_name:
                        name=package_name[:package_name.index(".")]
                    else:
                        name=package_name
                    if name in exceptional_names.keys():
                        name=exceptional_names[name]
                    name=get_custom_module_name(bl_type,name)
                    if name!='bl_ui' and name!='Brush_Manager':
                        registered_panels.append(bl_type)
                        
                        if  getattr(bl_type,'backup_space',None)==None:
                            bl_type.backup_space=getattr(bl_type,'bl_space_type',None)
                        if getattr(bl_type,'backup_region',None)==None:
                            bl_type.backup_region=getattr(bl_type,'bl_region_type',None)
                        if not getattr(bl_type,'backup_order',None)!=None:
                            bl_type.backup_order=getattr(bl_type,'bl_order',0)
                        if not getattr(bl_type,'backup_category',None) and getattr(bl_type,'bl_category',None):
                            bl_type.backup_category=bl_type.bl_category
                        # if not getattr(bl_type,'renamed_category',None): 
                        if getattr(bl_type,"bl_region_type","None")in {'HEADER','UI'} and getattr(bl_type,'bl_space_type',None) in {space_type,'TOPBAR'}:
                            
                            if getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)>=0 :
                                bl_type.renamed_category=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)].tab_name
            except:
                pass
        
        config_path=os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))), "config","CP-PanelOrder.txt")
        order_of_panels=[]
        if os.path.isfile(config_path):
            with open(config_path, mode='r', newline='\n', encoding='utf-8') as file:
                order_of_panels=file.readlines()
        else:
            if os.path.isfile(os.path.join(os.path.dirname(__file__),'CP-PanelOrder.txt')):
                with open(os.path.join(os.path.dirname(__file__),'CP-PanelOrder.txt'), mode='r', newline='\n', encoding='utf-8') as file:
                    order_of_panels=file.readlines()

        cleaned_order=[]
        if hasattr(bpy.utils,'panels'):
            if len(getattr(bpy.utils,'panels',[]))>len(order_of_panels):
                order_of_panels=getattr(bpy.utils,'panels',[])
        for o in order_of_panels:
            if o.replace("\n","") not in cleaned_order:
                cleaned_order.append(o.replace("\n",""))
        order_of_panels=cleaned_order
        
        focused_panels=[]
        
        if getattr(self,f"filter_enabled{get_active_space(space_type)}"):
            #context.workspace.use_filter_by_owner = True
            categories=[]
            for a in split_keep_substring(getattr(preferences(),f"addons_to_exclude{get_active_space(space_type)}"))+addons_to_exclude:
                try:
                    if get_module_name_from_addon_name(a)!='--Unknown--':
                        categories.append(get_module_name_from_addon_name(a))
                except:
                    pass
            for index,a in enumerate(getattr(preferences(),f"workspace_categories{get_active_space(space_type)}")):
                if getattr(self,f'enabled{get_active_space(space_type)}_{index}',False):
                    #if a.name==context.workspace.pap_active_workspace_category:
                        #categories_string= ''.join(a.panels.split())
                        categories_string=split_keep_substring(a.panels)
                        #categories.extend([a.strip() for a in categories_string])
                        categories.extend([get_module_name_from_addon_name(a) for a in categories_string])
            panels_from_reorder_list=[]
            for a in getattr(preferences(),f"addon_info{get_active_space(space_type)}"):
                for b in split_keep_substring(a.addons):
                    panels_from_reorder_list.append(b)
            panels_from_reorder_list=remove_duplicates(panels_from_reorder_list)
            if not preferences().sort_per_category:
                categories=sort_by_another_list(categories,panels_from_reorder_list)
            modules=[]
            for a in [__package__] + categories[:]:
                try:
                    if a in ['Tool','View','Node','Options']:
                         modules.append(a)
                    else:
                        modules.append(sys.modules[a].__name__)
                except:
                    pass
            panels_to_reregister=[]
            parents=[]
            children=[]
            for bl_type in registered_panels:
                try:
                    package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                    if "." in package_name:
                        name=package_name[:package_name.index(".")]
                    else:
                        name=package_name
                    if name in exceptional_names.keys():
                        name=exceptional_names[name]
                    name=get_custom_module_name(bl_type,name)
                    try:                  
                        if bl_type :
                            if (getattr(bl_type,'bl_category',None) and getattr(bl_type,'backup_space',None)==space_type and getattr(bl_type,"backup_region","None")=='UI' ) and not getattr(bl_type,'bl_parent_id',None):
                                package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                                
                                if "." in package_name:
                                    name=package_name[:package_name.index(".")]
                                else:
                                    name=package_name
                                if name in exceptional_names.keys():
                                    name=exceptional_names[name]
                                name=get_custom_module_name(bl_type,name)
                                if name in modules or name.replace('ender-','') in modules:
                                    if bl_type.bl_category!='Focused':
                                        bl_type.bl_category=bl_type.backup_category if getattr(bl_type,'backup_category',None) else bl_type.bl_category
                                    if bl_type.bl_category=='Focused':

                                        focused_panels.append((name,bl_type))
                                    else:
                                    #print(bl_type,getattr(bl_type,'backup_category',None))
                                        panels_to_reregister.append((name,bl_type))

                                    # bpy.utils.unregister_class(bl_type)  
                                    # bpy.utils.register_class(bl_type)
                            else:
                                if getattr(bl_type,'bl_parent_id',None) and getattr(bl_type,"backup_region","UI")=='UI':
                                    
                                    package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                                    if "." in package_name:
                                        name=package_name[:package_name.index(".")]
                                    else:
                                        name=package_name
                                    if name in exceptional_names.keys():
                                        name=exceptional_names[name]
                                    name=get_custom_module_name(bl_type,name)
                                    
                                    if name in modules or name.replace("ender-","") in modules:
                                        #print(bl_type.bl_category)
                                        #bl_type.bl_category=bl_type.backup_category if getattr(bl_type,'backup_category',None) else bl_type.bl_category
                                        #print(bl_type,getattr(bl_type,'backup_category',None))
                                        
                                        children.append((name,bl_type))
                                        parents.append(bl_type.bl_parent_id)
                                        # bpy.utils.unregister_class(bl_type)  
                                        # bpy.utils.register_class(bl_type)
                    except Exception as e:
                        pass
                        #print(e)
                except:
                    pass
            
            panels_to_reregister=sort_panels(panels_to_reregister,order_of_panels)
            children=sort_panels(children,order_of_panels)
            focused_panels=sort_panels(focused_panels,order_of_panels)
            
            # print([a for a in children if 'ac' in str(a[1]) and a[0]=='BlenRig-master'])
            children=sorted(children,key=lambda x:getattr(x[1],'bl_parent_id','Temp') in [getattr(a[1],'bl_idname','None') for a in children])
            # print([a for a in children if 'ac' in str(a[1]) and a[0]=='BlenRig-master'])
            
            for addon in modules:
                # print(addon)
                for name,p in focused_panels:

                    if name==addon or name.replace('ender-','')==addon:
                        if getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)>=0:
                            if getattr(p,'bl_category'):
                                if not getattr(p,'backup_category',None):
                                    p.backup_category=p.bl_category
                            if getattr(p,'backup_order',None)!=None:
                                p.bl_order=getattr(p,'backup_order')
                            if p.bl_category!='Focused':
                                # if name=='Animation_Layers':
                                #     print(p.bl_category,preferences().addon_info_for_renaming[preferences().addon_info_for_renaming.find(name)].tab_name)
                                p.bl_category=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)].tab_name
                                p.renamed_category=p.bl_category
                        try:
                            bpy.utils.unregister_class(p)  
                            register_panel(p)
                        except Exception as e:
                            print(e)
            for addon in modules:
                for name,p in panels_to_reregister:
                    if name==addon or name.replace('ender-','')==addon:
                        # print(name)
                        if getattr(p,'backup_space',None)!=None and getattr(p,'backup_region',None)!=None:
                            p.bl_space_type=p.backup_space
                            p.bl_region_type=p.backup_region
                        if "home_builder" in name:
                             p.bl_options={a for a in getattr(p,'bl_options',[]) if a!='HIDE_HEADER'}
                        if "pbr-painter" in name and getattr(p,'backup_category',None)=='Tool':
                            p.bl_category='Tool'
                            p.renamed_category='Tool'
                        elif name in addons_with_multiple_tabs.keys() and getattr(p,'backup_category',None)!=addons_with_multiple_tabs[name]:
                            
                            p.bl_category=getattr(p,'backup_category',None)
                            p.renamed_category=getattr(p,'backup_category',None)
                        # elif "cats-blender" in name and getattr(p,'backup_category',None)!='CATS':
                            
                        #     p.bl_category=getattr(p,'backup_category',None)
                        #     p.renamed_category=getattr(p,'backup_category',None)
                        # elif "shino" in name and getattr(p,'backup_category',None)!='Shino':
                            
                        #     p.bl_category=getattr(p,'backup_category',None)
                        #     p.renamed_category=getattr(p,'backup_category',None)
                        elif name in ['View','Tool','Node','Options']:
                            p.bl_category=getattr(p,'backup_category',None)
                            p.renamed_category=getattr(p,'backup_category',None)
                            p.bl_order=getattr(p,'backup_order',0)
                        else:
                            if getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)>=0:
                                if getattr(p,'bl_category'):
                                    
                                    if not getattr(p,'backup_category',None):
                                        p.backup_category=p.bl_category
                                if getattr(p,'backup_order',None)!=None:
                                    p.bl_order=getattr(p,'backup_order')
                                if p.bl_category!='Focused':
                                    p.bl_category=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)].tab_name
                                    p.renamed_category=p.bl_category
                        try:
                            bpy.utils.unregister_class(p)  
                            register_panel(p)
                        except Exception as e:
                            print(1,e)
            for name,c in children:
                if getattr(c,'backup_space',None)!=None and getattr(c,'backup_region',None)!=None:
                    c.bl_space_type=c.backup_space
                    c.bl_region_type=c.backup_region
                if "home_builder" in name:
                    c.bl_options={a for a in getattr(c,'bl_options',[]) if a!='HIDE_HEADER'}
                if "pbr-painter" in name and getattr(c,'backup_category',None)=='Tool':
                        c.bl_category='Tool'
                        c.renamed_category='Tool'
                elif name in addons_with_multiple_tabs.keys() and getattr(c,'backup_category',None)!=addons_with_multiple_tabs[name]:
                            
                            c.bl_category=getattr(c,'backup_category',None)
                            c.renamed_category=getattr(c,'backup_category',None)
                # elif "home_builder" in name and getattr(c,'backup_category',None)!='Home Builder':
                #         c.bl_category=getattr(c,'backup_category',None)
                #         c.renamed_category=getattr(c,'backup_category',None)
                # elif "cats-blender" in name and getattr(c,'backup_category',None)!='CATS':
                            
                #             c.bl_category=getattr(c,'backup_category',None)
                #             c.renamed_category=getattr(c,'backup_category',None)
                # elif "shino" in name and getattr(c,'backup_category',None)!='Shino':
                            
                #             c.bl_category=getattr(c,'backup_category',None)
                #             c.renamed_category=getattr(c,'backup_category',None)
                elif name in ['View','Tool','Node','Options']:
                            
                    c.bl_category=getattr(c,'backup_category',None)
                    c.renamed_category=getattr(c,'backup_category',None)
                    c.bl_order=getattr(c,'backup_order')
                else:
                    if getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)>=0:
                        if getattr(c,'bl_category',None):
                            if not getattr(c,'backup_category',None):
                                c.backup_category=c.bl_category
                        if getattr(c,'backup_order',None)!=None:
                                    c.bl_order=getattr(c,'backup_order')
                        c.bl_category=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)].tab_name
                        c.renamed_category=c.bl_category
                try:
                    if c.bl_parent_id in parents:
                        bpy.utils.unregister_class(c)  
                        register_panel(c)
                except Exception as e:
                            print(2,e)
            for name,c in children:
                if getattr(c,'backup_space',None)!=None and getattr(c,'backup_region',None)!=None:
                    c.bl_space_type=c.backup_space
                    c.bl_region_type=c.backup_region
                if "pbr-painter" in name and getattr(c,'backup_category',None)=='Tool':
                        c.bl_category='Tool'
                        c.renamed_category='Tool'
                elif name in addons_with_multiple_tabs.keys() and getattr(c,'backup_category',None)!=addons_with_multiple_tabs[name]:
                            
                    c.bl_category=getattr(c,'backup_category',None)
                    c.renamed_category=getattr(c,'backup_category',None)
                # elif "home_builder" in name and getattr(c,'backup_category',None)!='Home Builder':
                #         c.bl_category=getattr(c,'backup_category',None)
                #         c.renamed_category=getattr(c,'backup_category',None)
                # elif "cats-blender" in name and getattr(c,'backup_category',None)!='CATS':
                            
                #         c.bl_category=getattr(c,'backup_category',None)
                #         c.renamed_category=getattr(c,'backup_category',None)
                # elif "shino" in name and getattr(c,'backup_category',None)!='Shino':
                            
                #             c.bl_category=getattr(c,'backup_category',None)
                #             c.renamed_category=getattr(c,'backup_category',None)
                elif name in ['View','Tool','Node','Options']:
                            
                    c.bl_category=getattr(c,'backup_category',None)
                    c.renamed_category=getattr(c,'backup_category',None)
                    c.bl_order=getattr(c,'backup_order',0)
                else:
                    if getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)>=0:
                        if getattr(c,'bl_category',None):
                            if not getattr(c,'backup_category',None):
                                c.backup_category=c.bl_category
                        if getattr(c,'backup_order',None)!=None:
                                    c.bl_order=getattr(c,'backup_order')
                        c.bl_category=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)].tab_name
                        c.renamed_category=c.bl_category
                try:
                    if c.bl_parent_id not in parents:
                        bpy.utils.unregister_class(c)  
                        register_panel(c)
                except Exception as e:
                            print(3,e)
            
            #print([sys.modules[get_module_name_from_addon_name(a)] for a in preferences().addons_to_exclude)+addons_to_exclude])
            modules_to_remove=[]
            for b in bpy.context.preferences.addons.keys():
                try:
                    #print(b)
                    mod = sys.modules[b]
                    if mod.__name__ not in categories+[__package__]+["cycles",]:
                            a = mod.__name__
                            modules_to_remove.append(a)
                except:
                    pass
            for b in ['View','Tool','Node','Options']:
                try:
                    #print(b)
                    if b not in categories+[__package__]+["cycles",]:
                            a = b
                            modules_to_remove.append(a)
                except:
                    pass
                            #print("Marker",a)
            panels_to_unregister=[]
            children_to_unregister=[]
            parents_to_unregister=[]
            for bl_type in registered_panels:
                try:
                    if issubclass(bl_type, bpy.types.Panel):
                        if getattr(bl_type,"bl_parent_id","None")=='None' and getattr(bl_type,"backup_region","UI")=='UI' and getattr(bl_type,'backup_space',"VIEW_3D")==space_type :
                            
                            package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                            
                            if "." in package_name:
                                name=package_name[:package_name.index(".")]
                            else:
                                name=package_name
                            if name in exceptional_names.keys():
                                name=exceptional_names[name]
                            name=get_custom_module_name(bl_type,name)
                            if name in modules_to_remove or name.replace("ender-","") in modules_to_remove:
                                
                                if not getattr(bl_type,'backup_category',None):
                                    bl_type.backup_category=bl_type.bl_category if getattr(bl_type,'bl_category',preferences().holder_tab_name)!=preferences().holder_tab_name else (bl_type.backup_category if getattr(bl_type,'backup_category',None) else getattr(bl_type,'bl_category','None'))
                                if not getattr(bl_type,'backup_order',None)!=None:
                                    bl_type.backup_order=getattr(bl_type,'bl_order',0)
                                
                                if getattr(bl_type,'bl_category','None')!='Focused':
                                    bl_type.bl_category=preferences().holder_tab_name
                                    
                                    bl_type.bl_order=1000000
                                    if preferences().remove_holder_tab:
                                        bl_type.bl_region_type='HEADER'
                                        bl_type.bl_space_type='TOPBAR'
                                    panels_to_unregister.append((name,bl_type))
                                    
                                    
                            
                except:
                    pass
            for bl_type in registered_panels:
                try:
                    if issubclass(bl_type, bpy.types.Panel):
                        #print(getattr(bl_type,'bl_parent_id',None))
                        if getattr(bl_type,"bl_parent_id","None")!='None' and getattr(bl_type,"backup_region","UI")=='UI' and getattr(bl_type,'backup_space',"VIEW_3D")==space_type :
                            
                            package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                            
                            if "." in package_name:
                                name=package_name[:package_name.index(".")]
                            else:
                                name=package_name
                            if name in exceptional_names.keys():
                                name=exceptional_names[name]
                            name=get_custom_module_name(bl_type,name)
                            #print(name)
                            # print(modules_to_remove)
                            if name in modules_to_remove or name.replace("ender-","") in modules_to_remove:
                                #print(bl_type)
                                #print("296 setting backup",bl_type.bl_category)
                                
                                if not getattr(bl_type,'backup_category',None):
                                    bl_type.backup_category=bl_type.bl_category if getattr(bl_type,'bl_category',preferences().holder_tab_name)!=preferences().holder_tab_name else (bl_type.backup_category if getattr(bl_type,'backup_category',None) else getattr(bl_type,'bl_category','None'))
                                if not getattr(bl_type,'backup_order',None)!=None:
                                    bl_type.backup_order=getattr(bl_type,'bl_order',0)
                                
                                if getattr(bl_type,'bl_category','None')!='Focused' and getattr(getattr(bpy.types,bl_type.bl_parent_id,None),"bl_category","None")!='Focused':
                                    bl_type.bl_category=preferences().holder_tab_name
                                    
                                    bl_type.bl_order=1000000
                                    if preferences().remove_holder_tab:
                                        bl_type.bl_region_type='HEADER'
                                        bl_type.bl_space_type='TOPBAR'
                                    children_to_unregister.append((name,bl_type))
                                    parents_to_unregister.append(bl_type.bl_parent_id)
                                # if preferences().addon_info_for_renaming.find(name)>=0 and hasattr(bl_type,'bl_category'):
                                #     bl_type.renamed_category=preferences().addon_info_for_renaming[preferences().addon_info_for_renaming.find(name)].tab_name
                                #print("removing",bl_type)
                                    
                                    # bpy.utils.unregister_class(bl_type)  
                                    # if 'kit' in name.lower():
                                    #     print(bl_type,getattr(bl_type,'backup_space','None'))
                                    # if 'VIEW3D_PT_blenderkit_advanced_model_search' in getattr(bl_type,'bl_idname','None'):
                                    #     if getattr(bl_type,'bl_parent_id','None') in [getattr(a,'bl_idname','None') for a in registered_panels]:
                                    #         print(getattr(bl_type,'bl_parent_id','None'),"is registered")
                                    
                                    # try:
                                    #     register_panel(bl_type)
                                    #     # if 'VIEW3D_PT_blenderkit_unified' in getattr(bl_type,'bl_idname','None'):
                                    #     #     print("Registering",'VIEW3D_PT_blenderkit_unified')
                                    # except Exception as e:
                                    #     print(e)
                                    
                            
                except:
                    pass
            panels_to_unregister=sort_panels(panels_to_unregister,order_of_panels)
            children_to_unregister=sort_panels(children_to_unregister,order_of_panels)
            # print([a for a in panels_to_unregister if 'Sanctus' in a[0]])
            # print("\nC")
            # print([a for a in children_to_unregister if 'Sanctus' in a[0]])
            for _,bl_type in panels_to_unregister:
                bpy.utils.unregister_class(bl_type)  
                try:
                    register_panel(bl_type)
                except Exception as e:
                    print(e)
            for _,bl_type in children_to_unregister:
                if bl_type.bl_parent_id in parents_to_unregister:
                    bpy.utils.unregister_class(bl_type)  
                    try:
                        register_panel(bl_type)
                    except Exception as e:
                        print("Children Parent",e)
            for _,bl_type in children_to_unregister:
                if bl_type.bl_parent_id not in parents_to_unregister:
                    bpy.utils.unregister_class(bl_type)  
                    try:
                        register_panel(bl_type)
                    except Exception as e:
                        print("Children",e)
            
        else:
            # print("Disabled")
            # print(registered_panels)
            panels_to_reregister=[]
            parents=[]
            children=[]
            modules=[]
            focused_panels=[]
            for a in bpy.context.preferences.addons.keys():
                
                try:
                    if sys.modules[a].__name__ not in ["cycles",]:
                        modules.append(sys.modules[a].__name__)
                except Exception as e:
                    pass
            modules=['Tool','View','Node','Options']+modules
            sorted_modules=[]
            for addon_info in getattr(preferences(),f"addon_info{get_active_space(space_type)}"):
                for a in split_keep_substring(addon_info.addons):
                    if a in modules:
                        sorted_modules.append(a)
            sorted_modules=sorted_modules+[a for a in modules if a not in sorted_modules]
            for bl_type in registered_panels:
                    
                    try:
                            # if "RanTools" in str(bl_type):
                            #      print("ADASA",bl_type,getattr(bl_type,'bl_category',None), getattr(bl_type,'backup_space',None),space_type ,getattr(bl_type,'bl_parent_id',None))
                            if getattr(bl_type,'bl_category',None) and getattr(bl_type,'backup_space',None)==space_type and not getattr(bl_type,'bl_parent_id',None):
                                
                                package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                                
                                if "." in package_name:
                                    name=package_name[:package_name.index(".")]
                                else:
                                    name=package_name
                                if name in exceptional_names.keys():
                                        name=exceptional_names[name]
                                name=get_custom_module_name(bl_type,name)
                                #modules=[sys.modules[a].__name__ for a in bpy.context.preferences.addons.keys()]
                                if name in sorted_modules or name.replace('ender-','') in sorted_modules:
                                    #bl_type.backup_category=bl_type.bl_category
                                    if bl_type.bl_category!='Focused':
                                        bl_type.bl_category=bl_type.backup_category if getattr(bl_type,'backup_category',None) else bl_type.bl_category
                                    # bl_type.backup_category=None
                                    #print(bl_type,getattr(bl_type,'backup_category',None))
                                    # if preferences().addon_info_for_renaming.find(name)>=0:
                                    #     print("347 Setting backup",bl_type.bl_category)
                                    #     bl_type.backup_category=bl_type.bl_category
                                    #     #print(name,preferences().addon_info_for_renaming[preferences().addon_info_for_renaming.find(name)].tab_name)
                                    #     bl_type.bl_category=preferences().addon_info_for_renaming[preferences().addon_info_for_renaming.find(name)].tab_name
                                    if bl_type.bl_category=='Focused':
                                        focused_panels.append((name,bl_type))
                                    else:
                                        panels_to_reregister.append((name,bl_type))
                            else:
                                if bl_type and getattr(bl_type,'bl_parent_id',None) and getattr(bl_type,'backup_region','UI')=='UI':
                                    package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                                    if "." in package_name:
                                        name=package_name[:package_name.index(".")]
                                    else:
                                        name=package_name
                                    if name in exceptional_names.keys():
                                        name=exceptional_names[name]
                                    name=get_custom_module_name(bl_type,name)
                                    if name!='bl_ui' and name in sorted_modules:
                                        parents.append(bl_type.bl_parent_id)
                                        children.append((name,bl_type))
                    except Exception as e:
                        pass
            
            panels_to_reregister=sort_panels(panels_to_reregister,order_of_panels)
            children=sort_panels(children,order_of_panels)
            focused_panels=sort_panels(focused_panels,order_of_panels)
            children=sorted(children,key=lambda x:getattr(x[1],'bl_parent_id','Temp') in [getattr(a[1],'bl_idname','None') for a in children])
            #print(panels_to_reregister,children)
            # print("Sorted",[a.__name__ for b,a in panels_to_reregister if "RTOO" in a.__name__])
            for addon in sorted_modules:
                
                for name,p in focused_panels:
                    if getattr(p,'backup_space',None)!=None and getattr(p,'backup_region',None)!=None:
                        p.bl_space_type=p.backup_space
                        p.bl_region_type=p.backup_region
                    if getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)>=0 and hasattr(p,'bl_category'):
                        
                        if not getattr(p,'backup_category',None):
                            p.backup_category=p.bl_category
                        if getattr(p,'backup_order',None)!=None:
                                p.bl_order=getattr(p,'backup_order')
                        if p.bl_category!='Focused':
                            p.bl_category=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)].tab_name
                            p.renamed_category=p.bl_category
                    if name==addon or name.replace('ender-','')==addon:
                        try:
                            bpy.utils.unregister_class(p) 
                            register_panel(p)
                        except:
                            pass
            for addon in sorted_modules:
                for name,p in panels_to_reregister:
                    # print(p)
                    if getattr(p,'backup_space',None)!=None and getattr(p,'backup_region',None)!=None:
                        # print(getattr(p,'backup_region',None))
                        p.bl_space_type=p.backup_space
                        p.bl_region_type=p.backup_region
                    if "home_builder" in name:
                            p.bl_options={a for a in getattr(p,'bl_options',[]) if a!='HIDE_HEADER'}
                    if "pbr-painter" in name and getattr(p,'backup_category',None)=='Tool':
                        p.bl_category='Tool'
                        p.renamed_category='Tool'
                    elif name in addons_with_multiple_tabs.keys() and getattr(p,'backup_category',None)!=addons_with_multiple_tabs[name]:
                            
                            p.bl_category=getattr(p,'backup_category',None)
                            p.renamed_category=getattr(p,'backup_category',None)
                    # elif "home_builder" in name and getattr(p,'backup_category',None)!='Home Builder':
                            
                    #         p.bl_category=getattr(p,'backup_category',None)
                    #         p.renamed_category=getattr(p,'backup_category',None)
                    # elif "cats-blender" in name and getattr(p,'backup_category',None)!='CATS':
                            
                    #         p.bl_category=getattr(p,'backup_category',None)
                    #         p.renamed_category=getattr(p,'backup_category',None)
                    # elif "shino" in name and getattr(p,'backup_category',None)!='Shino':
                            
                    #         p.bl_category=getattr(p,'backup_category',None)
                    #         p.renamed_category=getattr(p,'backup_category',None)
                    elif name in ['View','Tool','Node','Options']:
                            
                            p.bl_category=getattr(p,'backup_category',None)
                            p.renamed_category=getattr(p,'backup_category',None)
                            p.bl_order=getattr(p,'backup_order',0)
                    else:
                        if getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)>=0 and hasattr(p,'bl_category'):
                            
                            if not getattr(p,'backup_category',None):
                                p.backup_category=p.bl_category
                            if getattr(p,'backup_order',None)!=None:
                                    p.bl_order=getattr(p,'backup_order')
                            if p.bl_category!='Focused':
                                p.bl_category=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)].tab_name
                                p.renamed_category=p.bl_category
                    if name==addon or name.replace('ender-','')==addon:
                        try:
                            
                            bpy.utils.unregister_class(p) 
                            register_panel(p)
                            # print(p.bl_region_type,p.bl_space_type)
                        except:
                            pass
            
            for name,c in children:
                if getattr(c,'backup_space',None)!=None and getattr(c,'backup_region',None)!=None:
                    c.bl_space_type=c.backup_space
                    c.bl_region_type=c.backup_region
                if "pbr-painter" in name and getattr(c,'backup_category',None)=='Tool':
                        c.bl_category='Tool'
                        c.renamed_category='Tool'
                elif name in addons_with_multiple_tabs.keys() and getattr(c,'backup_category',None)!=addons_with_multiple_tabs[name]:
                            
                            c.bl_category=getattr(c,'backup_category',None)
                            c.renamed_category=getattr(c,'backup_category',None)
                # elif "home_builder" in name and getattr(c,'backup_category',None)!='Home Builder':
                #         c.bl_category=getattr(c,'backup_category',None)
                #         c.renamed_category=getattr(c,'backup_category',None)
                # elif "cats-blender" in name and getattr(c,'backup_category',None)!='CATS':
                            
                #             c.bl_category=getattr(c,'backup_category',None)
                #             c.renamed_category=getattr(c,'backup_category',None)
                # elif "shino" in name and getattr(c,'backup_category',None)!='Shino':
                            
                #             c.bl_category=getattr(c,'backup_category',None)
                #             c.renamed_category=getattr(c,'backup_category',None)
                elif name in ['View','Tool','Node','Options']:
                            
                            c.bl_category=getattr(c,'backup_category',None)
                            c.renamed_category=getattr(c,'backup_category',None)
                            c.bl_order=getattr(c,'backup_order',0)
                else:
                    if getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)>=0 :
                        
                        if getattr(c,'bl_category',None) and not getattr(c,'backup_category',None):
                                c.backup_category=c.bl_category
                        if getattr(c,'backup_order',None)!=None:
                                    c.bl_order=getattr(c,'backup_order')
                        c.bl_category=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)].tab_name
                        c.renamed_category=c.bl_category
                try:
                    if c.bl_parent_id in parents:
                        bpy.utils.unregister_class(c)  
                        register_panel(c)
                except:
                    pass
            for name,c in children:
                if getattr(c,'backup_space',None)!=None and getattr(c,'backup_region',None)!=None:
                    c.bl_space_type=c.backup_space
                    c.bl_region_type=c.backup_region
                if "pbr-painter" in name and getattr(c,'backup_category',None)=='Tool':
                        c.bl_category='Tool'
                        c.renamed_category='Tool'
                elif name in addons_with_multiple_tabs.keys() and getattr(c,'backup_category',None)!=addons_with_multiple_tabs[name]:
                            
                            c.bl_category=getattr(c,'backup_category',None)
                            c.renamed_category=getattr(c,'backup_category',None)
                # elif "home_builder" in name and getattr(c,'backup_category',None)!='Home Builder':
                #         c.bl_category=getattr(c,'backup_category',None)
                #         c.renamed_category=getattr(c,'backup_category',None)
                # elif "cats-blender" in name and getattr(c,'backup_category',None)!='CATS':
                            
                #             c.bl_category=getattr(c,'backup_category',None)
                #             c.renamed_category=getattr(c,'backup_category',None)
                # elif "shino" in name and getattr(c,'backup_category',None)!='Shino':
                            
                #             c.bl_category=getattr(c,'backup_category',None)
                #             c.renamed_category=getattr(c,'backup_category',None)
                elif name in ['View','Tool','Node','Options']:
                            
                            c.bl_category=getattr(c,'backup_category',None)
                            c.renamed_category=getattr(c,'backup_category',None)
                            c.bl_order=getattr(c,'backup_order',0)
                else:
                    if getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)>=0 :
                        if getattr(c,'bl_category',None) and not getattr(c,'backup_category',None):
                                c.backup_category=c.bl_category
                        if getattr(c,'backup_order',None)!=None:
                                    c.bl_order=getattr(c,'backup_order')
                        c.bl_category=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space_type)}").find(name)].tab_name
                        c.renamed_category=c.bl_category
                try:
                    if c.bl_parent_id not in parents:
                        bpy.utils.unregister_class(c)  
                        register_panel(c)
                except:
                    pass
    context.workspace.enabled_categories=""
        
    for i in range(50):
        if getattr(preferences().categories,f"enabled{get_active_space(space_type)}_{i}",False):
            context.workspace.enabled_categories=context.workspace.enabled_categories+f"{i},"
    context.workspace.filter_enabled=preferences().categories.filter_enabled
def load_reordering_list(context,space='VIEW_3D'):
        for a in getattr(preferences(),f"addon_info{get_active_space(space)}"):
            a.addons=""
            a.ordered=""
        #preferences().addon_info.clear()
        change_panel_category("Turn OFF","Focused",space)
        workspace_category_enabled(preferences().categories,context) 
        og_filter=preferences().categories.filter_enabled
        preferences().categories.filter_enabled=False
        panels_in_use=[]
        no_fix_required=[]
        base_type = bpy.types.Panel
        for typename in dir(bpy.types):
            
            try:
                bl_type = getattr(bpy.types, typename,None)
                if issubclass(bl_type, base_type):
                    if getattr(bl_type,'bl_category',None) and getattr(bl_type,'bl_space_type',None)==space and getattr(bl_type,'bl_category',None) not in ["Item","Dev",preferences().holder_tab_name]:
                        panels_in_use.append(getattr(bl_type,'bl_category'))
                        if getattr(bl_type,'bl_category') not in [a.name for a in  getattr(preferences(),f"addon_info{get_active_space(space)}")]:
                            t=getattr(preferences(),f"addon_info{get_active_space(space)}").add()
                            t.name=getattr(bl_type,'bl_category')
                            package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                            if "." in package_name:
                                name=package_name[:package_name.index(".")]
                            else:
                                name=package_name
                            if name in exceptional_names.keys():
                                        name=exceptional_names[name]
                            name=get_custom_module_name(bl_type,name)
                            if name not in t.addons:
                                
                                t.addons=name
                            
                            if getattr(bl_type,'bl_order',None)!=None:
                                if getattr(bl_type,'bl_order',None)==0:
                                    no_fix_required.append(name)
                                t.ordered=name
                                #print(t.name,t.ordered)
                        else:
                            t=getattr(preferences(),f"addon_info{get_active_space(space)}")[getattr(preferences(),f"addon_info{get_active_space(space)}").find(getattr(bl_type,'bl_category'))]
                            package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                            if "." in package_name:
                                name=package_name[:package_name.index(".")]
                            else:
                                name=package_name
                            if name in exceptional_names.keys():
                                        name=exceptional_names[name]
                            name=get_custom_module_name(bl_type,name)
                            if name not in t.addons:
                                t.addons=t.addons+","+name
                            if getattr(bl_type,'bl_order',None) !=None:
                                if getattr(bl_type,'bl_order',None)==0:
                                    no_fix_required.append(name)
                                if name not in t.ordered:
                                    t.ordered=t.ordered+("," if t.ordered else '') +name
                                    #print(t.name,t.ordered)
            except Exception as e:
                if str(e)!="issubclass() arg 1 must be a class":
                    pass
        # print(no_fix_required)
        # for addon in preferences().addon_info:
        #     temp_array=addon.ordered)
        #     for a in no_fix_required:
        #         if a in temp_array:
        #             temp_array.remove(a)
        #     print(addon,temp_array)
        #     addon.ordered=",".join(temp_array)
        for addon in getattr(preferences(),f"addon_info{get_active_space(space)}"):
            

            if addon.ordered:
                if addon.ordered!=addon.addons:
                    temp_array=sorted(split_keep_substring(addon.addons),key=lambda x:x in addon.ordered)
                    addon.addons=",".join(temp_array)
                    addon.ordered=""
        # print(preferences().addon_info[:])
        for addon in getattr(preferences(),f"addon_info{get_active_space(space)}"):
            if addon.name not in panels_in_use:
                if getattr(preferences(),f"addon_info{get_active_space(space)}").find(addon.name)>-1:
                    getattr(preferences(),f"addon_info{get_active_space(space)}").remove(getattr(preferences(),f"addon_info{get_active_space(space)}").find(addon.name))
        # print(preferences().addon_info[:])
        # for addon in sorted([a.module for a in context.preferences.addons]):
        #     t=preferences().addon_info.add()
        #     t.name=addon
        # for b in preferences().addon_info:
        #     print(b.name,b.addons)
        savePreferences()
        preferences().categories.filter_enabled=og_filter
tab_names_dict={
     'pbr-pbr-painter':'PBR pbr-painter',
     'home_builder':'Home Builder',
     'shape_generator':'Shape Generator'
}
def load_renaming_list(context,reset=False,space='VIEW_3D'):
    change_panel_category("Turn OFF","Focused",space)
    workspace_category_enabled(preferences().categories,context) 
    og_filter=preferences().categories.filter_enabled
    preferences().categories.filter_enabled=False
    panels_in_use=[]
    no_fix_required=[]
    base_type = bpy.types.Panel
    for typename in dir(bpy.types):
        try:
            bl_type = getattr(bpy.types, typename,None)
            if issubclass(bl_type, base_type):
                if getattr(bl_type,'bl_category',None) and getattr(bl_type,'bl_space_type',None)==space:
                    
                    package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                    
                    if "." in package_name:
                        name=package_name[:package_name.index(".")]
                    else:
                        name=package_name
                    if name in exceptional_names.keys():
                        name=exceptional_names[name]
                    name=get_custom_module_name(bl_type,name)
                    if "bl_ui" not in name:
                        panels_in_use.append(name)
                        if name not in [a.name for a in  getattr(preferences(),f"addon_info_for_renaming{get_active_space(space)}")]:
                            
                            t=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space)}").add()
                               
                            t.tab_name=getattr(bl_type,'bl_category') if getattr(bl_type,'bl_category',preferences().holder_tab_name) !=preferences().holder_tab_name else (getattr(bl_type,'backup_category',preferences().holder_tab_name))
                            t.tab_name=tab_names_dict.get(name,t.tab_name)
                            t.name=name
                                #print(t.name,t.ordered)
                        else:
                            t=getattr(preferences(),f"addon_info_for_renaming{get_active_space(space)}")[getattr(preferences(),f"addon_info_for_renaming{get_active_space(space)}").find(name)]
                            if not t.tab_name or t.tab_name==preferences().holder_tab_name or t.tab_name=='None':
                                t.tab_name=getattr(bl_type,'bl_category') if getattr(bl_type,'bl_category',preferences().holder_tab_name) !=preferences().holder_tab_name else (getattr(bl_type,'backup_category',preferences().holder_tab_name))
                            if reset and getattr(bl_type,'backup_category',None):
                                 t.tab_name=getattr(bl_type,'backup_category')
                            # package_name=inspect.getmodule(bl_type).__package__ if inspect.getmodule(bl_type).__package__ else inspect.getmodule(bl_type).__name__
                            # if "." in package_name:
                            #     name=package_name[:package_name.index(".")]
                            # else:
                            #     name=package_name
                            # t.name=name
        except Exception as e:
            if str(e)!="issubclass() arg 1 must be a class":
                pass
    #print(panels_in_use)
    for a in getattr(preferences(),f"addon_info_for_renaming{get_active_space(space)}"):
        if a.name not in panels_in_use:
            getattr(preferences(),f"addon_info_for_renaming{get_active_space(space)}").remove(getattr(preferences(),f"addon_info_for_renaming{get_active_space(space)}").find(a.name))
    savePreferences()
    preferences().categories.filter_enabled=og_filter
def get_active_space(specified=None):
        if specified:
            space="_"+specified
            if space=='_VIEW_3D':
                space=''
            space=space.lower()
            return space
        space="_"+preferences().space_type
        if space=='_VIEW_3D':
            space=''
        space=space.lower()
        return space
def get_gist_text(id="",file="",all_files=False):
    try:
        try:
            r=requests.get('https://api.github.com/gists/' + id)
        except ConnectionError as e:
            print("Could not fetch gist! No Internet connection!")
            return None,'No Internet'
        results=r.json()
        if not all_files:
            if not file:
                file=list(results['files'].keys())[0]
            return str(results['files'][file]['content'])
        else:
            data={}
            for a in list(results['files'].keys()):
                data[a]=str(results['files'][a]['content'])
            return data
    except Exception as e:
        print("API Limit reached! Please try later!")
        return ''
def remove_duplicates_preserve_order(lst):
    seen = set()
    return [x for x in lst if not (x in seen or seen.add(x))]
def fetch_categories():
    import addon_utils
    enabled_addons=[addon_utils.module_bl_info(a)['name'] for a in addon_utils.modules() if addon_utils.check(a.__name__)[1]]
    
    categories=get_gist_text("377e6d64956df704765a71b10df6eeb6",all_files=True)
    for cat,addons in categories.items():
        addons=[a for a in split_keep_substring(addons) if a in enabled_addons]
        if ("(Node Editor)") in cat :
            if preferences().space_type=='NODE_EDITOR':
                if preferences().workspace_categories_node_editor.get(cat,None):
                    t=preferences().workspace_categories_node_editor.get(cat,None)
                    current_addons=split_keep_substring(preferences().workspace_categories_node_editor.get(cat).panels)
                    t.panels=",".join(remove_duplicates_preserve_order(current_addons+addons))
                else:
                    t=preferences().workspace_categories_node_editor.add()
                    t.name=cat
                    t.panels=",".join(addons)
        elif ("(Image Editor)") in cat :
            if preferences().space_type=='IMAGE_EDITOR':            
                if preferences().workspace_categories_image_editor.get(cat,None):
                    t=preferences().workspace_categories_image_editor.get(cat,None)
                    current_addons=split_keep_substring(preferences().workspace_categories_image_editor.get(cat).panels)
                    t.panels=",".join(remove_duplicates_preserve_order(current_addons+addons))
                else:
                    t=preferences().workspace_categories_image_editor.add()
                    t.name=cat
                    t.panels=",".join(addons)
        else:
            if preferences().space_type=='VIEW_3D':
                if preferences().workspace_categories.get(cat,None):
                    t=preferences().workspace_categories.get(cat,None)
                    current_addons=split_keep_substring(preferences().workspace_categories.get(cat).panels)
                    t.panels=",".join(remove_duplicates_preserve_order(current_addons+addons))
                else:
                    t=preferences().workspace_categories.add()
                    t.name=cat
                    t.panels=",".join(addons)
            